# -*- coding: utf-8 -*-
import abc
import datetime
import logging
import re

import pytz
from dateutil.parser import parse as date_parse

from .data_models import RangeGroupData, RangeGroupOperator
from .utils import FEELFunctionsManager
from .validators import BinaryOperationValidator, DummyValidator, ListsLengthValidator

logger = logging.getLogger(__name__)


class Expression(metaclass=abc.ABCMeta):
    validator_cls = DummyValidator

    @abc.abstractmethod
    def evaluate(self, context):
        pass


class CommonExpression(Expression):
    def __init__(self, value):
        self.value = value

    def evaluate(self, context):
        return self.value


class Expr(CommonExpression):
    def evaluate(self, context):
        return self.value.evaluate(context)


class Number(CommonExpression):
    pass


class String(CommonExpression):
    pass


class Boolean(CommonExpression):
    pass


class Null(Expression):
    def evaluate(self, context):
        return None


class List(Expression):
    def __init__(self, *items):
        self.items = items

    def evaluate(self, context):
        return [item.evaluate(context) for item in self.items]


class ListItem(Expression):
    def __init__(self, list_expr, index):
        self.list_expr = list_expr
        self.index = index

    def evaluate(self, context):
        items = self.list_expr.evaluate(context)
        if not isinstance(items, list) or self.index == 0 or len(items) < abs(self.index):
            return None
        items = items[self.index - 1] if self.index > 0 else items[self.index]
        return items


class ListMatch(Expression):
    validator_cls = ListsLengthValidator

    def __init__(self, iter_pairs, expr):
        self.iter_pairs = iter_pairs
        self.expr = expr

    def evaluate_and_validate_iter_pairs(self, context):
        iter_pairs = [(pair[0].value, pair[1].evaluate(context)) for pair in self.iter_pairs]
        self.validator_cls()(lists=[pair[1] for pair in iter_pairs])
        return iter_pairs


class ListEvery(ListMatch):
    def evaluate(self, context):
        iter_pairs = self.evaluate_and_validate_iter_pairs(context)
        for i in range(0, len(iter_pairs[0][1])):
            tmp_context = {**context, **{pair[0]: pair[1][i] for pair in iter_pairs}}
            if self.expr.evaluate(tmp_context) is False:
                return False
        return True


class ListSome(ListMatch):
    def evaluate(self, context):
        iter_pairs = self.evaluate_and_validate_iter_pairs(context)
        for i in range(0, len(iter_pairs[0][1])):
            tmp_context = {**context, **{pair[0]: pair[1][i] for pair in iter_pairs}}
            if self.expr.evaluate(tmp_context) is True:
                return True
        return False


class ListFilter(Expression):
    def __init__(self, list_expr, filter_expr):
        self.list_expr = list_expr
        self.filter_expr = filter_expr

    def evaluate(self, context):
        items = self.list_expr.evaluate(context)
        if not isinstance(items, list):
            return None
        result = []
        for item in items:
            try:
                # 当 item 为 dict 且 filter 中对比的 key 缺失时，可能报错
                if self.filter_expr.evaluate(item if isinstance(item, dict) else {"item": item}):
                    result.append(item)
            except Exception as e:
                logger.exception(e)
                pass
        return result


class Pair(Expression):
    def __init__(self, key, value):
        self.key = key
        self.value = value

    def evaluate(self, context):
        return self.key.evaluate(context), self.value.evaluate(context)


class Context(Expression):
    def __init__(self, pairs):
        self.pairs = pairs

    def evaluate(self, context):
        return dict(pair.evaluate(context) for pair in self.pairs)


class ContextItem(Expression):
    def __init__(self, expr, keys):
        self.expr = expr
        self.keys = keys

    def evaluate(self, context):
        result = self.expr.evaluate(context)
        for key in self.keys:
            if not isinstance(result, dict):
                return None
            result = result.get(key)
        return result


class Variable(Expression):
    def __init__(self, name):
        self.name = name

    def evaluate(self, context):
        return context.get(self.name)


class FunctionCall(Expression):
    def __init__(self, name, args):
        self.name = name
        self.args = args

    def evaluate(self, context):
        function = context.get(self.name)
        if function is None:
            raise ValueError(f"Unknown function: {self.name}")
        return function(*[arg.evaluate(context) for arg in self.args])


class BinaryOperator(Expression):
    def __init__(self, left, right):
        self.left = left
        self.right = right


class SameTypeBinaryOperator(BinaryOperator):
    validator_cls = BinaryOperationValidator

    def __init__(self, operation, left, right):
        super().__init__(left, right)
        self.operation = operation

    def evaluate(self, context):
        left_val = self.left.evaluate(context)
        right_val = self.right.evaluate(context)
        self.validator_cls()(left_val, right_val)
        return getattr(self, self.operation)(left_val, right_val)

    def add(self, left_val, right_val):
        return left_val + right_val

    def subtract(self, left_val, right_val):
        return left_val - right_val

    def multiply(self, left_val, right_val):
        return left_val * right_val

    def divide(self, left_val, right_val):
        return left_val / right_val

    def power(self, left_val, right_val):
        return left_val**right_val

    def equal(self, left_val, right_val):
        return left_val == right_val

    def less_than(self, left_val, right_val):
        return left_val < right_val

    def greater_than(self, left_val, right_val):
        return left_val > right_val

    def less_than_or_equal(self, left_val, right_val):
        return left_val <= right_val

    def greater_than_or_equal(self, left_val, right_val):
        return left_val >= right_val


class NotEqual(BinaryOperator):
    def evaluate(self, context):
        return self.left.evaluate(context) != self.right.evaluate(context)


class And(BinaryOperator):
    def evaluate(self, context):
        return self.left.evaluate(context) and self.right.evaluate(context)


class Or(BinaryOperator):
    def evaluate(self, context):
        return self.left.evaluate(context) or self.right.evaluate(context)


class In(BinaryOperator):
    def evaluate(self, context):
        left_val = self.left.evaluate(context)
        right_val = self.right.evaluate(context)
        if isinstance(self.right, RangeGroup):
            left_operation = (
                left_val > right_val.left_val
                if right_val.left_operator == RangeGroupOperator.GT
                else left_val >= right_val.left_val
            )
            right_operation = (
                left_val < right_val.right_val
                if right_val.right_operator == RangeGroupOperator.LT
                else left_val <= right_val.right_val
            )
            return left_operation and right_operation
        return left_val in right_val


class Between(Expression):
    def __init__(self, value, left, right):
        self.value = value
        self.min = left
        self.max = right

    def evaluate(self, context):
        value = self.value.evaluate(context)
        return self.min.evaluate(context) <= value <= self.max.evaluate(context)


class RangeGroup(BinaryOperator):
    def __init__(self, left, right, left_operator, right_operator):
        self.left = left
        self.right = right
        self.left_operator = left_operator
        self.right_operator = right_operator

    def evaluate(self, context):
        left_val = self.left.evaluate(context)
        right_val = self.right.evaluate(context)
        data = {
            "left_val": left_val,
            "right_val": right_val,
            "left_operator": self.left_operator,
            "right_operator": self.right_operator,
        }
        return RangeGroupData(**data)


class BeforeFunc(BinaryOperator):
    def evaluate(self, context):
        left_val = self.left.evaluate(context)
        right_val = self.right.evaluate(context)
        operator = None
        if isinstance(self.left, RangeGroup):
            if left_val.right_operator == RangeGroupOperator.LT:
                operator = RangeGroupOperator.GTE
            left_val = left_val.right_val
        if isinstance(self.right, RangeGroup):
            if right_val.left_operator == RangeGroupOperator.GT:
                operator = RangeGroupOperator.GTE
            right_val = right_val.left_val

        if operator == RangeGroupOperator.GTE:
            return left_val <= right_val

        return left_val < right_val


class AfterFunc(BinaryOperator):
    def evaluate(self, context):
        left_val = self.left.evaluate(context)
        right_val = self.right.evaluate(context)
        operator = None
        if isinstance(self.left, RangeGroup):
            if left_val.left_operator == RangeGroupOperator.GT:
                operator = RangeGroupOperator.GTE
            left_val = left_val.left_val
        if isinstance(self.right, RangeGroup):
            if right_val.right_operator == RangeGroupOperator.LT:
                operator = RangeGroupOperator.GTE
            right_val = right_val.right_val
        if operator == RangeGroupOperator.GTE:
            return left_val >= right_val
        return left_val > right_val


class IncludesFunc(BinaryOperator):
    def evaluate(self, context):
        left_val: RangeGroupData = self.left.evaluate(context)
        right_val = self.right.evaluate(context)
        if isinstance(self.right, RangeGroup):
            left_operation = left_val.left_val <= right_val.left_val
            if left_val.left_operator == RangeGroupOperator.GT and right_val.left_operator == RangeGroupOperator.GTE:
                left_operation = left_val.left_val < right_val.left_val
            right_operation = left_val.right_val >= right_val.right_val
            if left_val.right_operator == RangeGroupOperator.LT and right_val.right_operator == RangeGroupOperator.LTE:
                right_operation = left_val.right_val > right_val.right_val
        else:
            left_operation = left_val.left_val <= right_val
            if left_val.left_operator == RangeGroupOperator.GT:
                left_operation = left_val.left_val < right_val
            right_operation = left_val.right_val >= right_val
            if left_val.right_operator == RangeGroupOperator.LT:
                right_operation = left_val.right_val > right_val
        return left_operation and right_operation


class Not(Expression):
    def __init__(self, value):
        self.value = value

    def evaluate(self, context):
        return not self.value.evaluate(context)


class Date(CommonExpression):
    def evaluate(self, context):
        year, month, day = self.value.split("-")
        return datetime.date(int(year), int(month), int(day))


class TZInfo(Expression):
    def __init__(self, method, value):
        self.method = method
        self.value = value

    def evaluate(self, context):
        if self.method == "name":
            return pytz.timezone(self.value)
        elif self.method == "offset":
            hours, minutes = map(int, self.value.split(":"))
            sign = -1 if hours < 0 else 1
            hours = abs(hours)
            offset = hours * 60 + minutes
            return pytz.FixedOffset(sign * offset)


class Time(Expression):
    def __init__(self, value, timezone: TZInfo = None):
        self.value = value
        self.timezone = timezone

    def evaluate(self, context):
        parsed_dt = date_parse(self.value)
        timezone = self.timezone.evaluate(context) if self.timezone is not None else None
        return datetime.time(parsed_dt.hour, parsed_dt.minute, parsed_dt.second, tzinfo=timezone)


class DateAndTime(Expression):
    def __init__(self, date: Date, time: Time):
        self.date = date
        self.time = time

    def evaluate(self, context):
        date = self.date.evaluate(context)
        time = self.time.evaluate(context)
        return datetime.datetime.combine(date, time, tzinfo=time.tzinfo)


class NowFunc(Expression):
    def evaluate(self, context):
        # TODO：带时区需要配置
        return datetime.datetime.now()


class TodayFunc(Expression):
    def evaluate(self, context):
        return datetime.date.today()


class DayOfWeekFunc(CommonExpression):
    WEEKDAYS = [
        "Monday",
        "Tuesday",
        "Wednesday",
        "Thursday",
        "Friday",
        "Saturday",
        "Sunday",
    ]

    def evaluate(self, context):
        date_or_datetime = self.value.evaluate(context)
        return self.WEEKDAYS[date_or_datetime.weekday()]


class MonthOfYearFunc(CommonExpression):
    MONTH_MAPPING = {
        1: "January",
        2: "February",
        3: "March",
        4: "April",
        5: "May",
        6: "June",
        7: "July",
        8: "Auguest",
        9: "September",
        10: "October",
        11: "November",
        12: "December",
    }

    def evaluate(self, context):
        date_or_datetime = self.value.evaluate(context)
        return self.MONTH_MAPPING[date_or_datetime.month]


class ToString(CommonExpression):
    def evaluate(self, context):
        return str(self.value.evaluate(context))


class StringOperator(BinaryOperator):
    validator_cls = BinaryOperationValidator

    def __init__(self, operation, left, right):
        super().__init__(left, right)
        self.operation = operation

    def evaluate(self, context):
        left_val = self.left.evaluate(context)
        right_val = self.right.evaluate(context)
        self.validator_cls()(left_val, right_val, instance_type=str)
        return getattr(self, self.operation)(left_val, right_val)

    def contains(self, left_str, right_str):
        return right_str in left_str

    def starts_with(self, left_str, right_str):
        return left_str.startswith(right_str)

    def ends_with(self, left_str, right_str):
        return left_str.endswith(right_str)

    def matches(self, left_str, right_str):
        return re.match(right_str, left_str) is not None


class ListOperator(Expression):
    def __init__(self, operation, *expr):
        self.operation = operation
        self.expr = expr

    def evaluate(self, context):
        return getattr(self, self.operation)(context)

    def list_contains(self, context):
        list_ = self.expr[0].evaluate(context)
        item = self.expr[1].evaluate(context)
        return item in list_

    def list_count(self, context):
        list_ = self.expr[0].evaluate(context)
        return len(list_)

    def list_all(self, context):
        list_ = self.expr[0].evaluate(context)
        return all(list_)

    def list_any(self, context):
        list_ = self.expr[0].evaluate(context)
        return any(list_)


class FuncInvocation(Expression):
    def __init__(self, func_name, args=None, named_args=None):
        self.func_name = func_name
        self.args = args or []
        self.named_args = named_args or {}

    def evaluate(self, context):
        try:
            func = FEELFunctionsManager.get_func(self.func_name)
        except Exception as e:
            logger.exception(e)
            func = None
        if not func:
            return None

        if self.args:
            params = [arg.evaluate(context) for arg in self.args]
            return func(*params)
        elif self.named_args:
            params = {key: arg.evaluate(context) for key, arg in self.named_args.items()}
            return func(**params)

        return func()



# -*- coding: utf-8 -*-
import abc
import importlib
import logging
from typing import Callable

from pydantic import BaseModel
from pydantic import ValidationError as PydanticValidationError
from pydantic import validate_arguments

from bkflow_feel.exceptions import ValidationError

logger = logging.getLogger(__name__)


class FEELFunctionsManager:
    __hub = {}

    @classmethod
    def register_invocation_cls(cls, invocation_cls):
        func_name = invocation_cls.Meta.func_name
        existed_invocation_cls = cls.__hub.get(func_name)
        if existed_invocation_cls:
            raise RuntimeError(
                "func register error, {}'s func_name {} conflict with {}".format(
                    existed_invocation_cls, func_name, invocation_cls
                )
            )

        cls.__hub[func_name] = invocation_cls

    @classmethod
    def register_funcs(cls, func_dict):
        for func_name, func_path in func_dict.items():
            if not isinstance(func_name, str):
                raise ValueError(f"func_name {func_name} should be string")
            if func_name in cls.__hub:
                raise ValueError(
                    "func register error, {}'s func_name {} conflict with {}".format(
                        func_path, func_name, cls.__hub[func_name]
                    )
                )
            cls.__hub[func_name] = func_path

    @classmethod
    def clear(cls):
        cls.__hub = {}

    @classmethod
    def all_funcs(cls):
        funcs = {}
        for version, invocation_cls in cls.__hub.items():
            funcs[version] = invocation_cls
        return funcs

    @classmethod
    def get_func(cls, func_name) -> Callable:
        func_obj = cls.__hub.get(func_name)
        if not func_obj:
            raise ValueError("func object {} not found".format(func_name))

        if isinstance(func_obj, FEELInvocationMeta):
            return func_obj()
        else:
            module_path, func_name = str(func_obj).rsplit(".", 1)
            module = importlib.import_module(module_path)
            func = getattr(module, func_name)
            return func

    @classmethod
    def func_call(cls, func_name, *args, **kwargs):
        func = cls.get_func(func_name)
        return func(*args, **kwargs)


class FEELInvocationMeta(type):
    """
    Metaclass for FEEL function invocation
    """

    def __new__(cls, name, bases, dct):
        # ensure initialization is only performed for subclasses of Plugin
        parents = [b for b in bases if isinstance(b, FEELInvocationMeta)]
        if not parents:
            return super().__new__(cls, name, bases, dct)

        new_cls = super().__new__(cls, name, bases, dct)

        # meta validation
        meta_obj = getattr(new_cls, "Meta", None)
        if not meta_obj:
            raise AttributeError("Meta class is required")

        func_name = getattr(meta_obj, "func_name", None)
        if not func_name:
            raise AttributeError("func_name is required in Meta")

        desc = getattr(meta_obj, "desc", None)
        if desc is not None and not isinstance(desc, str):
            raise AttributeError("desc in Meta should be str")

        # register func
        FEELFunctionsManager.register_invocation_cls(new_cls)

        return new_cls


class InvocationInputsModel(BaseModel):
    """
    Model for inputs of invocation
    """

    pass


class BaseFEELInvocation(metaclass=FEELInvocationMeta):
    """
    Base class for FEEL function invocation
    """

    class Inputs(InvocationInputsModel):
        """
        Inputs class for FEEL function invocation
        """

        pass

    @validate_arguments
    def __call__(self, *args, **kwargs):

        # 输入参数校验, 仅可能是 args 或 kwargs 之一
        try:
            params = {}
            if args:
                inputs_meta = getattr(self.Inputs, "Meta", None)
                inputs_ordering = getattr(inputs_meta, "ordering", None)
                if isinstance(inputs_ordering, list):
                    if len(args) > len(inputs_ordering):
                        raise ValidationError(f"Too many arguments for inputs: {args}")
                    params = {k: v for k, v in zip(inputs_ordering, args)}
            elif kwargs:
                params = kwargs

            if params:
                self.Inputs(**params)
        except PydanticValidationError as e:
            raise ValidationError(e)

        return self.invoke(*args, **kwargs)

    @abc.abstractmethod
    def invoke(self, *args, **kwargs):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
import abc

from .exceptions import ValidationError


class ValidationResult:
    def __init__(self, result: bool, err_msg: str = None, *args, **kwargs):
        self.result = result
        self.err_msg = err_msg


class Validator(metaclass=abc.ABCMeta):
    def __call__(self, *args, **kwargs):
        validation_result = self.validate(*args, **kwargs)
        if not validation_result.result:
            raise ValidationError(validation_result.err_msg)
        return validation_result

    @abc.abstractmethod
    def validate(self, *args, **kwargs) -> ValidationResult:
        pass


class DummyValidator(Validator):
    def validate(self, *args, **kwargs):
        pass


class ListsLengthValidator(Validator):
    def validate(self, lists, *args, **kwargs):
        if not lists or all(len(alist) == len(lists[0]) for alist in lists):
            return ValidationResult(True)
        return ValidationResult(False, "lists length not equal")


class BinaryOperationValidator(Validator):
    def validate(self, left_item, right_item, instance_type=None, *args, **kwargs) -> ValidationResult:
        if not isinstance(left_item, type(right_item)):
            return ValidationResult(
                False, f"Type of both operators must be same, get {type(left_item)} and {type(right_item)}",
            )
        if instance_type is not None and not isinstance(left_item, instance_type):
            return ValidationResult(
                False, f"Type of both operators must be {instance_type}, get {type(left_item)} and {type(right_item)}",
            )
        return ValidationResult(True)



# -*- coding: utf-8 -*-
import logging

from . import parser as default_parser
from . import transformer as default_transformer
from .exceptions import ValidationError
from .parsers import Expression

logger = logging.getLogger(__name__)


def parse_expression(
    expression, context=None, raise_exception=True, parser=default_parser, transformer=default_transformer,
):
    parse_tree = parser.parse(expression)
    logger.debug(parse_tree)
    ast = transformer.transform(parse_tree)
    logger.debug(ast)
    if not isinstance(ast, Expression):
        msg = f"Invalid FEEL expression: {expression}, ast: {ast}"
        logger.error(msg)
        if raise_exception:
            raise ValueError(msg)
        return None
    try:
        result = ast.evaluate(context or {})
    except ValidationError as e:
        logger.exception(f"evaluate expression error: {e}")
        if raise_exception:
            raise e
        return None
    except Exception as e:
        logger.exception(f"evaluate expression error: {e}")
        if raise_exception:
            raise e
        return None
    return result



# -*- coding: utf-8 -*-
from lark import Token, Transformer, v_args

from .data_models import RangeGroupOperator
from .parsers import (
    AfterFunc,
    And,
    BeforeFunc,
    Between,
    Boolean,
    Context,
    ContextItem,
    Date,
    DateAndTime,
    DayOfWeekFunc,
    Expr,
    FuncInvocation,
    FunctionCall,
    In,
    IncludesFunc,
    List,
    ListEvery,
    ListFilter,
    ListItem,
    ListOperator,
    ListSome,
    MonthOfYearFunc,
    Not,
    NotEqual,
    NowFunc,
    Null,
    Number,
    Or,
    Pair,
    RangeGroup,
    SameTypeBinaryOperator,
    String,
    StringOperator,
    Time,
    TodayFunc,
    ToString,
    TZInfo,
    Variable,
)


@v_args(inline=True)
class FEELTransformer(Transformer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def number(self, token):
        try:
            value = int(token.value)
        except ValueError:
            value = float(token.value)
        return Number(value)

    def string(self, value):
        return String(value[1:-1])

    def true(self):
        return Boolean(True)

    def false(self):
        return Boolean(False)

    def null(self):
        return Null()

    def expr(self, item):
        return Expr(item)

    def list_(self, *items):
        return List(*items)

    def list_filter(self, list_expr, filter_expr):
        return ListFilter(list_expr, filter_expr)

    def list_item(self, list_expr, index):
        return ListItem(list_expr=list_expr, index=int(index.value))

    def list_some(self, *items):
        expr = items[-1]
        iter_pairs = [(items[i], items[i + 1]) for i in range(0, len(items) - 1, 2)]
        return ListSome(iter_pairs, expr)

    def list_every(self, *items):
        expr = items[-1]
        iter_pairs = [(items[i], items[i + 1]) for i in range(0, len(items) - 1, 2)]
        return ListEvery(iter_pairs, expr)

    def range_atom(self, value):
        if isinstance(value, Token):
            if value.type == "SIGNED_INT":
                return Number(int(value.value))
            elif value.type == "NAME":
                return Variable(value.value)
        return value

    def close_range_group(self, start, end):
        return RangeGroup(start, end, RangeGroupOperator.GTE, RangeGroupOperator.LTE)

    def open_range_group(self, start, end):
        return RangeGroup(start, end, RangeGroupOperator.GT, RangeGroupOperator.LT)

    def left_open_range_group(self, start, end):
        return RangeGroup(start, end, RangeGroupOperator.GT, RangeGroupOperator.LTE)

    def right_open_range_group(self, start, end):
        return RangeGroup(start, end, RangeGroupOperator.GTE, RangeGroupOperator.LT)

    def variable(self, name_token):
        return Variable(name_token.value)

    def function_call(self, name, *args):
        return FunctionCall(name, args)

    def add(self, left, right):
        return SameTypeBinaryOperator("add", left, right)

    def sub(self, left, right):
        return SameTypeBinaryOperator("subtract", left, right)

    def mul(self, left, right):
        return SameTypeBinaryOperator("multiply", left, right)

    def div(self, left, right):
        return SameTypeBinaryOperator("divide", left, right)

    def pow(self, left, right):
        return SameTypeBinaryOperator("power", left, right)

    def eq(self, left, right):
        return SameTypeBinaryOperator("equal", left, right)

    def ne(self, left, right):
        return NotEqual(left, right)

    def lt(self, left, right):
        return SameTypeBinaryOperator("less_than", left, right)

    def gt(self, left, right):
        return SameTypeBinaryOperator("greater_than", left, right)

    def le(self, left, right):
        return SameTypeBinaryOperator("less_than_or_equal", left, right)

    def ge(self, left, right):
        return SameTypeBinaryOperator("greater_than_or_equal", left, right)

    def and_(self, left, right):
        return And(left, right)

    def or_(self, left, right):
        return Or(left, right)

    def between(self, target, left, right):
        return Between(target, left, right)

    def in_(self, left, right):
        return In(left, right)

    def not_(self, value):
        return Not(value)

    def date(self, value):
        return Date(value)

    def date_func(self, value):
        return Expr(value)

    def timezone(self, value):
        return Expr(value)

    def tz_offset(self, token):
        return TZInfo("offset", token.value)

    def tz_name(self, *tokens):
        if len(tokens) == 2:
            name = f"{tokens[0].value}/{tokens[1].value}"
        else:
            name = "UTC"
        return TZInfo("name", name)

    def time(self, value, timezone=None):
        return Time(value, timezone)

    def date_and_time(self, date, time):
        return DateAndTime(date, time)

    def time_func(self, value):
        return Expr(value)

    def date_and_time_func(self, value):
        return Expr(value)

    def now_func(self):
        return NowFunc()

    def today_func(self):
        return TodayFunc()

    def day_of_week_func(self, value):
        return DayOfWeekFunc(value)

    def month_of_year_func(self, value):
        return MonthOfYearFunc(value)

    def before_func(self, left, right):
        return BeforeFunc(left, right)

    def after_func(self, left, right):
        return AfterFunc(left, right)

    def includes_func(self, left, right):
        return IncludesFunc(left, right)

    def context(self, *args):
        return Context(args)

    def context_item(self, expr, *keys):
        return ContextItem(expr, keys)

    def pair(self, key_token, value):
        return Pair(key=String(key_token.value.strip('"')), value=value)

    def bracket(self, value):
        return Expr(value)

    def to_string(self, value):
        return ToString(value)

    def contains(self, left, right):
        return StringOperator("contains", left, right)

    def starts_with(self, left, right):
        return StringOperator("starts_with", left, right)

    def ends_with(self, left, right):
        return StringOperator("ends_with", left, right)

    def matches(self, left, right):
        return StringOperator("matches", left, right)

    def list_contains(self, *expr):
        return ListOperator("list_contains", *expr)

    def list_count(self, *expr):
        return ListOperator("list_count", *expr)

    def list_all(self, *expr):
        return ListOperator("list_all", *expr)

    def list_any(self, *expr):
        return ListOperator("list_any", *expr)

    def func_invocation(self, func_name, *args):
        func_name = " ".join([token.value for token in func_name.children])
        if len(args) == 1 and getattr(args[0], "data") and args[0].data.value == "named_args":
            args_pairs = args[0].children
            named_args = {args_pairs[i].value: args_pairs[i + 1] for i in range(0, len(args_pairs), 2)}
            return FuncInvocation(func_name, named_args=named_args)
        return FuncInvocation(func_name, args=args)



# -*- coding: utf-8 -*-
import enum
from typing import Any

from pydantic import BaseModel


class RangeGroupOperator(enum.Enum):
    GT = "greater than"
    GTE = "greater than or equal"
    LT = "less than"
    LTE = "less than or equal"


class RangeGroupData(BaseModel):
    left_val: Any
    right_val: Any
    left_operator: RangeGroupOperator
    right_operator: RangeGroupOperator
